"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var util_1 = require("./util");
function defaultCompare(a, b) {
    if (a < b) {
        return -1;
    }
    else if (a > b) {
        return 1;
    }
    else {
        return 0;
    }
}
var SortedOp;
(function (SortedOp) {
    SortedOp[SortedOp["NONE"] = -1] = "NONE";
    SortedOp[SortedOp["SKIP"] = 0] = "SKIP";
    SortedOp[SortedOp["PATCH"] = 1] = "PATCH";
    SortedOp[SortedOp["INSERT"] = 2] = "INSERT";
    SortedOp[SortedOp["INSERT_IDENTITY"] = 3] = "INSERT_IDENTITY";
    SortedOp[SortedOp["COPY"] = 4] = "COPY";
})(SortedOp || (SortedOp = {}));
var MuSortedArray = (function () {
    function MuSortedArray(schema, compare, identity) {
        this.muType = 'sorted-set';
        this.pool = [];
        this.muData = schema;
        this.compare = compare || defaultCompare;
        if (identity) {
            var x = identity.slice();
            x.sort(compare);
            for (var i = 0; i < x.length; ++i) {
                x[i] = schema.clone(x[i]);
            }
            this.identity = x;
        }
        else {
            this.identity = [];
        }
        this.json = {
            type: 'sorted-set',
            valueType: schema.json,
            identity: JSON.stringify(this.identity),
        };
    }
    MuSortedArray.prototype.alloc = function () {
        return this.pool.pop() || [];
    };
    MuSortedArray.prototype.free = function (set) {
        var schema = this.muData;
        for (var i = 0; i < set.length; ++i) {
            schema.free(set[i]);
        }
        set.length = 0;
        this.pool.push(set);
    };
    MuSortedArray.prototype.equal = function (x, y) {
        if (!Array.isArray(x) || !Array.isArray(y)) {
            return false;
        }
        if (x.length !== y.length) {
            return false;
        }
        for (var i = x.length - 1; i >= 0; --i) {
            if (!this.muData.equal(x[i], y[i])) {
                return false;
            }
        }
        return true;
    };
    MuSortedArray.prototype.clone = function (set) {
        var schema = this.muData;
        var result = this.alloc();
        result.length = set.length;
        for (var i = 0; i < set.length; ++i) {
            result[i] = schema.clone(set[i]);
        }
        return result;
    };
    MuSortedArray.prototype.copy = function (source, target) {
        if (source === target) {
            return;
        }
        var sourceLength = source.length;
        var targetLength = target.length;
        for (var i = sourceLength; i < targetLength; ++i) {
            this.muData.free(target[i]);
        }
        target.length = sourceLength;
        if (util_1.isMuPrimitive(this.muData.muType)) {
            for (var i = 0; i < sourceLength; ++i) {
                target[i] = source[i];
            }
            return;
        }
        for (var i = targetLength; i < target.length; ++i) {
            target[i] = this.muData.clone(source[i]);
        }
        for (var i = 0; i < Math.min(sourceLength, targetLength); ++i) {
            this.muData.copy(source[i], target[i]);
        }
    };
    MuSortedArray.prototype.diff = function (base, target, out) {
        if (base.length === 0 && target.length === 0) {
            return false;
        }
        var schema = this.muData;
        var compare = this.compare;
        out.grow(8);
        var head = out.offset;
        var opPtr = head;
        var opCount = 0;
        var opCode = SortedOp.NONE;
        out.offset += 4;
        var numOps = 0;
        function emitOp() {
            if (opCount > 0) {
                out.writeUint32At(opPtr, (opCount << 3) | opCode);
                numOps++;
            }
            out.grow(4);
            opPtr = out.offset;
            out.offset += 4;
        }
        var basePtr = 0;
        var targetPtr = 0;
        while (basePtr < base.length && targetPtr < target.length) {
            var baseItem = base[basePtr];
            var targetItem = target[targetPtr];
            var cmp = compare(baseItem, targetItem);
            if (cmp < 0) {
                if (opCode !== SortedOp.SKIP) {
                    emitOp();
                    opCount = 1;
                    opCode = SortedOp.SKIP;
                }
                else {
                    opCount++;
                }
                basePtr++;
            }
            else if (0 < cmp) {
                if (opCode === SortedOp.INSERT) {
                    if (schema.diff(schema.identity, targetItem, out)) {
                        opCount++;
                    }
                    else {
                        emitOp();
                        opCode = SortedOp.INSERT_IDENTITY;
                        opCount = 1;
                    }
                }
                else if (opCode === SortedOp.INSERT_IDENTITY) {
                    var prevOffset = out.offset;
                    out.grow(4);
                    out.offset += 4;
                    if (schema.diff(schema.identity, targetItem, out)) {
                        emitOp();
                        out.offset -= 4;
                        opPtr = prevOffset;
                        opCode = SortedOp.INSERT;
                        opCount = 1;
                    }
                    else {
                        out.offset -= 4;
                        opCount += 1;
                    }
                }
                else {
                    emitOp();
                    opCount = 1;
                    if (schema.diff(schema.identity, targetItem, out)) {
                        opCode = SortedOp.INSERT;
                    }
                    else {
                        opCode = SortedOp.INSERT_IDENTITY;
                    }
                }
                targetPtr++;
            }
            else {
                if (opCode === SortedOp.PATCH) {
                    if (schema.diff(baseItem, targetItem, out)) {
                        opCount++;
                    }
                    else {
                        emitOp();
                        opCode = SortedOp.COPY;
                        opCount = 1;
                    }
                }
                else if (opCode === SortedOp.COPY) {
                    var prevOffset = out.offset;
                    out.grow(4);
                    out.offset += 4;
                    if (schema.diff(baseItem, targetItem, out)) {
                        emitOp();
                        out.offset -= 4;
                        opPtr = prevOffset;
                        opCode = SortedOp.PATCH;
                        opCount = 1;
                    }
                    else {
                        out.offset -= 4;
                        opCount += 1;
                    }
                }
                else {
                    emitOp();
                    opCount = 1;
                    if (schema.diff(baseItem, targetItem, out)) {
                        opCode = SortedOp.PATCH;
                    }
                    else {
                        opCode = SortedOp.COPY;
                    }
                }
                basePtr++;
                targetPtr++;
            }
        }
        if (basePtr < base.length) {
            if (opCode !== SortedOp.SKIP) {
                emitOp();
                opCount = base.length - basePtr;
                opCode = SortedOp.SKIP;
            }
            else {
                opCount += base.length - basePtr;
            }
            basePtr++;
        }
        while (targetPtr < target.length) {
            var targetItem = target[targetPtr];
            if (opCode === SortedOp.INSERT) {
                if (schema.diff(schema.identity, targetItem, out)) {
                    opCount++;
                }
                else {
                    emitOp();
                    opCode = SortedOp.INSERT_IDENTITY;
                    opCount = 1;
                }
            }
            else if (opCode === SortedOp.INSERT_IDENTITY) {
                var prevOffset = out.offset;
                out.grow(4);
                out.offset += 4;
                if (schema.diff(schema.identity, targetItem, out)) {
                    emitOp();
                    out.offset -= 4;
                    opPtr = prevOffset;
                    opCode = SortedOp.INSERT;
                    opCount = 1;
                }
                else {
                    out.offset -= 4;
                    opCount += 1;
                }
            }
            else {
                emitOp();
                opCount = 1;
                if (schema.diff(schema.identity, targetItem, out)) {
                    opCode = SortedOp.INSERT;
                }
                else {
                    opCode = SortedOp.INSERT_IDENTITY;
                }
            }
            targetPtr++;
        }
        if (numOps === 0 && opCode === SortedOp.COPY && opCount === base.length) {
            out.offset = head;
            return false;
        }
        if (opCode !== SortedOp.SKIP) {
            emitOp();
        }
        out.offset -= 4;
        out.writeUint32At(head, numOps);
        return true;
    };
    MuSortedArray.prototype.patch = function (base, inp) {
        var schema = this.muData;
        var result = this.alloc();
        var opCount = inp.readUint32();
        var ptr = 0;
        for (var i = 0; i < opCount; ++i) {
            var code = inp.readUint32();
            var count = code >> 3;
            var op = code & 0x7;
            switch (op) {
                case SortedOp.INSERT_IDENTITY:
                    for (var j = 0; j < count; ++j) {
                        result.push(schema.clone(schema.identity));
                    }
                    break;
                case SortedOp.INSERT:
                    for (var j = 0; j < count; ++j) {
                        result.push(schema.patch(schema.identity, inp));
                    }
                    break;
                case SortedOp.PATCH:
                    for (var j = 0; j < count; ++j) {
                        result.push(schema.patch(base[ptr++], inp));
                    }
                    break;
                case SortedOp.COPY:
                    for (var j = 0; j < count; ++j) {
                        result.push(schema.clone(base[ptr++]));
                    }
                    break;
                case SortedOp.SKIP:
                    ptr += count;
                    break;
            }
        }
        return result;
    };
    return MuSortedArray;
}());
exports.MuSortedArray = MuSortedArray;
//# sourceMappingURL=sorted.js.map