"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var test = require("tape");
var mustreams_1 = require("mustreams");
var __1 = require("../");
var constants_1 = require("../constants");
var helper_1 = require("./helper");
test('struct - muData', function (t) {
    var spec = {
        v: new __1.MuFloat32(),
    };
    var structSchema = new __1.MuStruct(spec);
    t.notEquals(structSchema.muData, spec);
    t.equals(structSchema.muData.v, spec.v);
    t.end();
});
test('struct - identity', function (t) {
    var structSchema = new __1.MuStruct({
        v: new __1.MuFloat64(0.233),
        vs: new __1.MuVector(new __1.MuFloat64(0.233), 2),
        s: new __1.MuString('foo'),
        b: new __1.MuBoolean(),
    });
    t.equals(structSchema.identity.v, 0.233);
    t.same(structSchema.identity.vs, [0.233, 0.233]);
    t.equals(structSchema.identity.s, 'foo');
    t.equals(structSchema.identity.b, false);
    t.end();
});
test('struct - equal()', function (t) {
    var structSchema = new __1.MuStruct({
        b: new __1.MuBoolean(),
        s: new __1.MuStruct({
            b: new __1.MuBoolean(),
            s: new __1.MuStruct({
                b: new __1.MuBoolean(),
            }),
        }),
    });
    var a = structSchema.alloc();
    var b = structSchema.alloc();
    t.ok(structSchema.equal(a, b));
    b.s.s.b = !b.s.s.b;
    t.notOk(structSchema.equal(a, b));
    t.end();
});
test('struct - copy()', function (t) {
    var structSchema = new __1.MuStruct({
        s: new __1.MuStruct({
            s: new __1.MuStruct({
                a: new __1.MuASCII(),
                b: new __1.MuBoolean(),
                fa: new __1.MuFixedASCII(5),
                f32: new __1.MuFloat32(),
                f64: new __1.MuFloat64(),
                i8: new __1.MuInt8(),
                i16: new __1.MuInt16(),
                i32: new __1.MuInt32(),
                str: new __1.MuString(),
                u8: new __1.MuUint8(),
                u16: new __1.MuUint16(),
                u32: new __1.MuUint32(),
            }),
        }),
    });
    var source = structSchema.alloc();
    var target = structSchema.alloc();
    source.s.s.a = helper_1.randomValue('ascii');
    source.s.s.b = true;
    source.s.s.fa = 'abcde';
    source.s.s.f32 = helper_1.randomValue('float32');
    source.s.s.f64 = helper_1.randomValue('float64');
    source.s.s.i8 = helper_1.randomValue('int8');
    source.s.s.i16 = helper_1.randomValue('int16');
    source.s.s.i32 = helper_1.randomValue('int32');
    source.s.s.str = helper_1.randomValue('string');
    source.s.s.u8 = helper_1.randomValue('uint8');
    source.s.s.u16 = helper_1.randomValue('uint16');
    source.s.s.u32 = helper_1.randomValue('uint32');
    structSchema.copy(source, target);
    t.deepEqual(target, source);
    t.end();
});
test('struct (flat) - diff() & patch()', function (t) {
    function structSpec() {
        var result = {};
        for (var _i = 0, muPrimitiveTypes_1 = constants_1.muPrimitiveTypes; _i < muPrimitiveTypes_1.length; _i++) {
            var muType = muPrimitiveTypes_1[_i];
            var valueSchema = helper_1.muPrimitiveSchema(muType);
            if (valueSchema) {
                result[helper_1.randomString()] = valueSchema;
            }
        }
        return result;
    }
    var _loop_1 = function (i) {
        var spec = structSpec();
        var structSchema = new __1.MuStruct(spec);
        var testPatchingPair = helper_1.testPatchingPairFactory(t, structSchema);
        var randomStruct = function () {
            var result = {};
            var propNames = Object.keys(spec);
            var muTypes = propNames.map(function (name) { return spec[name].muType; });
            propNames.forEach(function (propName, idx) {
                result[propName] = helper_1.randomValue(muTypes[idx]);
            });
            return result;
        };
        testPatchingPair(randomStruct(), randomStruct());
    };
    for (var i = 0; i < 500; ++i) {
        _loop_1(i);
    }
    t.end();
});
test('struct (nested)', function (t) {
    function structSchemaOf(depth) {
        if (depth <= 2) {
            return new __1.MuStruct({
                type: new __1.MuString('nested'),
                struct: new __1.MuStruct({
                    type: new __1.MuString('flat'),
                }),
            });
        }
        return new __1.MuStruct({
            type: new __1.MuString('nested'),
            struct: structSchemaOf(depth - 1),
        });
    }
    function modifyStruct(s) {
        if (s.struct) {
            s.type = 'branch';
            modifyStruct(s.struct);
        }
        else {
            s.type = 'leaf';
        }
    }
    for (var depth = 2; depth < 100; ++depth) {
        var structSchema = structSchemaOf(depth);
        var identity = structSchema.identity;
        var struct = structSchema.clone(identity);
        t.notEquals(struct, identity);
        t.same(struct, identity);
        var ws = new mustreams_1.MuWriteStream(2);
        structSchema.diff(identity, struct, ws);
        var rs = new mustreams_1.MuReadStream(ws.buffer.uint8);
        t.same(structSchema.patch(identity, rs), struct);
        modifyStruct(struct);
        var testPatchingPair = helper_1.testPatchingPairFactory(t, structSchema);
        testPatchingPair(identity, struct);
    }
    t.end();
});
test('random test', function (t) {
    var testSchema = new __1.MuStruct({
        a: new __1.MuFloat32(1),
        b: new __1.MuUint8(),
        c: new __1.MuVector(new __1.MuFloat32(), 3),
        foo: new __1.MuString('0'),
    });
    function randomStruct() {
        var result = testSchema.alloc();
        result.a = ((Math.random() * 2) | 0) / 2;
        result.b = (Math.random() * 2) | 0;
        result.c = new Float32Array([
            ((Math.random() * 2) | 0) / 2,
            ((Math.random() * 2) | 0) / 2,
            ((Math.random() * 2) | 0) / 2
        ]);
        result.foo = Math.round(Math.random() * 8).toString();
        return result;
    }
    function calcDiff(a, b) {
        var x = new mustreams_1.MuWriteStream(1);
        testSchema.diff(a, b, x);
        return new mustreams_1.MuReadStream(x.bytes());
    }
    function testPair(a, b) {
        var x = calcDiff(a, b);
        if (x.length === 0) {
            t.same(a, b, 'structs are equal');
        }
        else {
            var c = testSchema.patch(a, x);
            t.same(b, c, 'patch ok');
            t.equals(x.length, x.offset, 'offset ok');
        }
    }
    for (var i = 0; i < 100; ++i) {
        var as = randomStruct();
        var bs = randomStruct();
        testPair(as, bs);
        testPair(bs, as);
        testPair(testSchema.identity, as);
        testPair(testSchema.identity, bs);
    }
    t.end();
});
//# sourceMappingURL=struct.js.map